/*
 * CastRegistry.hpp
 *
 *  Created on: 11. 7. 2017
 *	  Author: Jan Travnicek
 */

#ifndef _CAST_REGISTRY_HPP_
#define _CAST_REGISTRY_HPP_

#include <memory>
#include <string>
#include <set>
#include <map>

#include <exception/CommonException.h>
#include <abstraction/OperationAbstraction.hpp>

namespace abstraction {

class CastRegistry {
	class Entry {
		bool m_normalize;
	public:
		Entry ( bool normalize ) : m_normalize ( normalize ) {
		}

		virtual std::shared_ptr < abstraction::OperationAbstraction > getAbstraction ( ) const = 0;

		bool getNormalize ( ) const {
			return m_normalize;
		}

	};

	template < class Return, class Param >
	class DefaultEntryImpl : public Entry {
	public:
		DefaultEntryImpl ( bool normalize ) : Entry ( normalize ) {
		}

		virtual std::shared_ptr < abstraction::OperationAbstraction > getAbstraction ( ) const override;
	};

	template < class Return, class Param >
	class AlgorithmEntryImpl : public Entry {
		std::function < Return ( Param ) > m_callback;
	public:
		AlgorithmEntryImpl ( std::function < Return ( Param ) > callback, bool normalize ) : Entry ( normalize ), m_callback ( callback ) {
		}

		virtual std::shared_ptr < abstraction::OperationAbstraction > getAbstraction ( ) const override;
	};

	static ext::map < std::pair < std::string, std::string >, std::unique_ptr < Entry > > & getEntries ( ) {
		static ext::map < std::pair < std::string, std::string >, std::unique_ptr < Entry > > casts;
		return casts;
	};

public:
	template < class TargetType, class ParamType >
	static void registerCast ( std::string target, std::string param, bool normalize ) {
		if ( ! getEntries ( ).insert ( std::make_pair ( std::make_pair ( target, param ), std::unique_ptr < Entry > ( new DefaultEntryImpl < TargetType, ParamType > ( normalize ) ) ) ).second )
			throw ::exception::CommonException ( "Entry from " + param + " to " + target + " already registered." );
	}

	template < class TargetType, class ParamType >
	static void registerCast ( bool normalize ) {
		std::string target = ext::to_string < TargetType > ( );
		std::string param = ext::to_string < ParamType > ( );

		registerCast < TargetType, ParamType > ( target, param, normalize );
	}

	template < class TargetType, class ParamType >
	static void registerCastAlgorithm ( std::string target, std::string param, TargetType ( * callback ) ( const ParamType & ), bool normalize ) {
		if ( ! getEntries ( ).insert ( std::make_pair ( std::make_pair ( target, param ), std::unique_ptr < Entry > ( new AlgorithmEntryImpl < TargetType, const ParamType & > ( callback, normalize ) ) ) ).second )
			throw ::exception::CommonException ( "Entry from " + param + " to " + target + " already registered." );
	}

	template < class TargetType, class ParamType >
	static void registerCastAlgorithm ( TargetType ( * callback ) ( const ParamType & ), bool normalize ) {
		std::string target = ext::to_string < TargetType > ( );
		std::string param = ext::to_string < ParamType > ( );

		registerCastAlgorithm < TargetType, ParamType > ( target, param, callback, normalize );
	}

	template < class TargetType, class ParamType >
	static void registerCastAlgorithm ( std::string target, std::string param, TargetType ( * callback ) ( ParamType ), bool normalize ) {
		if ( ! getEntries ( ).insert ( std::make_pair ( std::make_pair ( target, param ), std::unique_ptr < Entry > ( new AlgorithmEntryImpl < TargetType, ParamType > ( callback, normalize ) ) ) ).second )
			throw ::exception::CommonException ( "Entry from " + param + " to " + target + " already registered." );
	}

	template < class TargetType, class ParamType >
	static void registerCastAlgorithm ( TargetType ( * callback ) ( ParamType ), bool normalize ) {
		std::string target = ext::to_string < TargetType > ( );
		std::string param = ext::to_string < ParamType > ( );

		registerCastAlgorithm < TargetType, ParamType > ( target, param, callback, normalize );
	}

	static std::shared_ptr < abstraction::OperationAbstraction > getAbstraction ( const std::string & target, const std::string & param, bool & normalize );

	static bool isNoOp ( const std::string & target, const std::string & param );

	static bool castAvailable ( const std::string & target, const std::string & param );

	static std::shared_ptr < abstraction::OperationAbstraction > getAbstraction ( const std::string & target, const std::string & param );

	static ext::set < std::string > listFrom ( const std::string & type );

	static ext::set < std::string > listTo ( const std::string & type );

	static ext::set < ext::pair < std::string, std::string > > list ( );
};

} /* namespace abstraction */

#include <abstraction/CastAbstraction.hpp>
#include <abstraction/AlgorithmAbstraction.hpp>

namespace abstraction {

template < class Return, class Param >
std::shared_ptr < abstraction::OperationAbstraction > CastRegistry::DefaultEntryImpl < Return, Param >::getAbstraction ( ) const {
	return std::make_shared < abstraction::CastAbstraction < Return, const Param & > > ( );
}

template < class Return, class Param >
std::shared_ptr < abstraction::OperationAbstraction > CastRegistry::AlgorithmEntryImpl < Return, Param >::getAbstraction ( ) const {
	return std::make_shared < abstraction::AlgorithmAbstraction < Return, const Param & > > ( m_callback );
}

} /* namespace abstraction */

#endif /* _CAST_REGISTRY_HPP_ */
