/*
 * Deque.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_DEQUE_H_
#define OBJECTS_DEQUE_H_

#include <deque>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>
#include <object/UniqueObject.h>
#include <object/AnyObject.h>

#include <core/normalize.hpp>

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occurred.
 */
template < class ElementType >
class ObjectsDeque final : public ext::deque < ElementType >, public ContainerBase {
public:
	explicit ObjectsDeque ( ext::deque < ElementType > );
	explicit ObjectsDeque ( );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsDeque & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef ObjectsDeque < > normalized_type;
};

template < class ElementType >
ObjectsDeque < ElementType >::ObjectsDeque ( ext::deque < ElementType > raw ) : ext::deque < ElementType > ( std::move ( raw ) ) {
}

template < class ElementType >
ObjectsDeque < ElementType >::ObjectsDeque ( ) : ext::deque < ElementType > ( ) {
}

template < class ElementType >
ContainerBase * ObjectsDeque < ElementType >::clone ( ) const {
	return new ObjectsDeque ( * this );
}

template < class ElementType >
ContainerBase * ObjectsDeque < ElementType >::plunder ( ) && {
	return new ObjectsDeque ( std::move ( * this ) );
}

template < class ElementType >
int ObjectsDeque < ElementType >::compare ( const ObjectsDeque & other ) const {
	static ext::compare < ext::deque < ElementType > > comp;

	return comp ( static_cast < const ext::deque < ElementType > & > ( * this ), static_cast < const ext::deque < ElementType > & > ( other ) );
}

template < class ElementType >
void ObjectsDeque < ElementType >::operator >>( std::ostream & os ) const {
	os << "(ObjectsDeque " << static_cast < const ext::deque < ElementType > & > ( * this ) << ")";
}

template < class ElementType >
ObjectsDeque < ElementType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return std::move ( ss ).str ( );
}

template < class ElementType >
alib::ObjectBase* ObjectsDeque < ElementType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace alib {

template < typename T >
struct xmlApi < ext::deque < T > > {
	static ext::deque < T > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const ext::deque < T > & data );
};

template < typename T >
ext::deque < T > xmlApi < ext::deque < T > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );

	ext::deque < T > set;

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		set.push_back ( alib::xmlApi < T >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );
	return set;
}

template < typename T >
bool xmlApi < ext::deque < T > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < typename T >
const std::string & xmlApi < ext::deque < T > >::xmlTagName ( ) {
	static std::string xmlTagName = "Deque";

	return xmlTagName;
}

template < typename T >
void xmlApi < ext::deque < T > >::compose ( ext::deque < sax::Token > & output, const ext::deque < T > & input ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );

	for ( const T & item : input )
		alib::xmlApi < T >::compose ( output, item );

	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

template < typename T >
struct xmlApi < container::ObjectsDeque < T > > {
	static container::ObjectsDeque < T > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const container::ObjectsDeque < T > & data );
};

template < typename T >
container::ObjectsDeque < T > xmlApi < container::ObjectsDeque < T > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	return container::ObjectsDeque < T > ( xmlApi < ext::deque < T > >::parse ( input ) );
}

template < typename T >
bool xmlApi < container::ObjectsDeque < T > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return xmlApi < ext::deque < T > >::first ( input );
}

template < typename T >
const std::string & xmlApi < container::ObjectsDeque < T > >::xmlTagName ( ) {
	return xmlApi < ext::deque < T > >::xmlTagName ( );
}

template < typename T >
void xmlApi < container::ObjectsDeque < T > >::compose ( ext::deque < sax::Token > & output, const container::ObjectsDeque < T > & input ) {
	return xmlApi < ext::deque < T > >::compose ( output, input );
}

template < >
struct normalize < ext::deque < alib::Object > > {
	static ext::deque < alib::Object > && eval ( ext::deque < alib::Object > && source ) {
		return std::move ( source );
	}
};

template < class ElementType >
struct normalize < ext::deque < ElementType > > {
	static ext::deque < alib::Object > eval ( ext::deque < ElementType > && source ) {
		ext::deque < alib::Object > res;

		for ( ElementType & element : source )
			res.push_back ( alib::Object ( alib::AnyObject < ElementType > ( std::move ( element ) ) ) );

		return res;
	}
};

template < class ElementType >
struct normalize < container::ObjectsDeque < ElementType >, typename std::enable_if < ! std::is_same < container::ObjectsDeque < ElementType >, container::ObjectsDeque < > >::value >::type > {
	static container::ObjectsDeque < > eval ( container::ObjectsDeque < ElementType > && value ) {
		return container::ObjectsDeque < > ( normalize < ext::deque < ElementType > >::eval ( std::move ( value ) ) );
	}
};

} /* namespace alib */

#endif /* OBJECTS_DEQUE_H_ */
