/*
 * Map.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_MAP_H_
#define OBJECTS_MAP_H_

#include <map>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>
#include <object/UniqueObject.h>
#include <object/AnyObject.h>

#include "ObjectsPair.h"

#include <core/normalize.hpp>

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occured.
 */
template < class KeyType, class ValueType >
class ObjectsMap final : public ext::map < KeyType, ValueType >, public ContainerBase {
public:
	explicit ObjectsMap ( ext::map < KeyType, ValueType > );
	explicit ObjectsMap ( );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsMap & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef ObjectsMap < > normalized_type;
};

template < class KeyType, class ValueType >
ObjectsMap<KeyType, ValueType>::ObjectsMap ( ext::map < KeyType, ValueType > raw ) : ext::map < KeyType, ValueType > ( std::move ( raw ) ) {

}

template < class KeyType, class ValueType >
ObjectsMap<KeyType, ValueType>::ObjectsMap() : ext::map<KeyType, ValueType>() {

}

template < class KeyType, class ValueType >
ContainerBase* ObjectsMap<KeyType, ValueType>::clone() const {
	return new ObjectsMap(*this);
}

template < class KeyType, class ValueType >
ContainerBase* ObjectsMap<KeyType, ValueType>::plunder() && {
	return new ObjectsMap(std::move(*this));
}

template < class KeyType, class ValueType >
int ObjectsMap<KeyType, ValueType>::compare(const ObjectsMap& other) const {
	static ext::compare<ext::map<KeyType, ValueType>> comp;
	return comp ( static_cast < const ext::map < KeyType, ValueType > & > ( * this ), static_cast < const ext::map < KeyType, ValueType > & > ( other ) );
}

template < class KeyType, class ValueType >
void ObjectsMap<KeyType, ValueType>::operator>>(std::ostream& os) const {
	os << "(ObjectsMap " << static_cast < const ext::map < KeyType, ValueType > & > ( * this ) << ")";
}

template < class KeyType, class ValueType >
ObjectsMap<KeyType, ValueType>::operator std::string() const {
	std::stringstream ss;
	ss << *this;
	return std::move(ss).str();
}

template < class KeyType, class ValueType >
alib::ObjectBase* ObjectsMap < KeyType, ValueType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace alib {

template < typename T, typename R >
struct xmlApi < ext::map < T, R > > {
	static ext::map < T, R > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const ext::map < T, R > & data );
};

template < typename T, typename R >
ext::map < T, R > xmlApi < ext::map < T, R > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );

	ext::map < T, R > map;

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) )
		map.insert ( alib::xmlApi < ext::pair < T, R > >::parse ( input ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );
	return map;
}

template < typename T, typename R >
bool xmlApi < ext::map < T, R > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < typename T, typename R >
const std::string & xmlApi < ext::map < T, R > >::xmlTagName ( ) {
	static std::string xmlTagName = "Map";

	return xmlTagName;
}

template < typename T, typename R >
void xmlApi < ext::map < T, R > >::compose ( ext::deque < sax::Token > & output, const ext::map < T, R > & input ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );

	for ( const std::pair < const T, R > & item : input )
		alib::xmlApi < ext::pair < const T, R > >::compose ( output, item );

	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

template < typename T, typename R >
struct xmlApi < container::ObjectsMap < T, R > > {
	static container::ObjectsMap < T, R > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const container::ObjectsMap < T, R > & data );
};

template < typename T, typename R >
container::ObjectsMap < T, R > xmlApi < container::ObjectsMap < T, R > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	return container::ObjectsMap < T, R > ( xmlApi < ext::map < T, R > >::parse ( input ) );
}

template < typename T, typename R >
bool xmlApi < container::ObjectsMap < T, R > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < typename T, typename R >
const std::string & xmlApi < container::ObjectsMap < T, R > >::xmlTagName ( ) {
	return xmlApi < ext::map < T, R > >::xmlTagName ( );
}

template < typename T, typename R >
void xmlApi < container::ObjectsMap < T, R > >::compose ( ext::deque < sax::Token > & output, const container::ObjectsMap < T, R > & input ) {
	xmlApi < ext::map < T, R > >::compose ( output, input );
}

template < >
struct normalize < ext::map < alib::Object, alib::Object > > {
	static ext::map < alib::Object, alib::Object > && eval ( ext::map < alib::Object, alib::Object > && source ) {
		return std::move ( source );
	}
};

template < class KeyType, class ValueType >
struct normalize < ext::map < KeyType, ValueType > > {
	static ext::map < alib::Object, alib::Object > eval ( ext::map < KeyType, ValueType > && source ) {
		ext::map < alib::Object, alib::Object > res;

		for ( std::pair < KeyType, ValueType > && element : ext::make_moveable_map ( source ) )
			res.insert ( normalize < std::pair < KeyType, ValueType > >::eval ( std::move ( element ) ) );

		return res;
	}
};

template < class KeyType, class ValueType >
struct normalize < container::ObjectsMap < KeyType, ValueType >, typename std::enable_if < ! std::is_same < container::ObjectsMap < KeyType, ValueType >, container::ObjectsMap < > >::value >::type > {
	static container::ObjectsMap < > eval ( container::ObjectsMap < KeyType, ValueType > && value ) {
		return container::ObjectsMap < > ( normalize < ext::map < KeyType, ValueType > >::eval ( std::move ( value ) ) );
	}
};

} /* namespace alib */

#endif /* OBJECTS_MAP_H_ */
