/*
 * Pair.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_PAIR_H_
#define OBJECTS_PAIR_H_

#include <pair>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>
#include <object/UniqueObject.h>
#include <object/AnyObject.h>

#include <core/normalize.hpp>

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occured.
 */
template < class FirstType, class SecondType >
class ObjectsPair final : public ext::pair < FirstType, SecondType >, public ContainerBase {
public:
	explicit ObjectsPair ( FirstType firstObject, SecondType secondObject );
	explicit ObjectsPair ( ext::pair < FirstType, SecondType > );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsPair & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef ObjectsPair < > normalized_type;
};

template < class FirstType, class SecondType >
ObjectsPair < FirstType, SecondType >::ObjectsPair(ext::pair<FirstType, SecondType> raw) : ext::pair<FirstType, SecondType>(std::move(raw)) {

}

template < class FirstType, class SecondType >
ObjectsPair < FirstType, SecondType >::ObjectsPair(FirstType firstObject, SecondType secondObject) : ext::pair<FirstType, SecondType>(std::move(firstObject), std::move(secondObject)) {

}

template < class FirstType, class SecondType >
ContainerBase* ObjectsPair < FirstType, SecondType >::clone() const {
	return new ObjectsPair(*this);
}

template < class FirstType, class SecondType >
ContainerBase* ObjectsPair < FirstType, SecondType >::plunder() && {
	return new ObjectsPair(std::move(*this));
}

template < class FirstType, class SecondType >
int ObjectsPair < FirstType, SecondType >::compare(const ObjectsPair& other) const {
	static ext::compare<ext::pair<FirstType, SecondType>> comp;
	return comp ( static_cast < const ext::pair < FirstType, SecondType > & > ( * this ), static_cast < const ext::pair < FirstType, SecondType > & > ( other ) );
}

template < class FirstType, class SecondType >
void ObjectsPair < FirstType, SecondType >::operator>>(std::ostream& os) const {
	os << "(ObjectsPair " << static_cast < const ext::pair < FirstType, SecondType > & > ( * this ) << ")";
}

template < class FirstType, class SecondType >
ObjectsPair < FirstType, SecondType >::operator std::string() const {
	std::stringstream ss;
	ss << *this;
	return std::move(ss).str();
}

template < class FirstType, class SecondType >
alib::ObjectBase* ObjectsPair < FirstType, SecondType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace alib {

template < typename T, typename R >
struct xmlApi < ext::pair < T, R > > {
	static ext::pair < T, R > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const ext::pair < T, R > & data );
	static void compose ( ext::deque < sax::Token > & output, const std::pair < T, R > & data );
};

template < typename T, typename R >
ext::pair < T, R > xmlApi < ext::pair < T, R > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );

	T first = alib::xmlApi < T >::parse ( input );
	R second = alib::xmlApi < R >::parse ( input );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );

	return ext::make_pair ( std::move ( first ), std::move ( second ) );
}

template < typename T, typename R >
bool xmlApi < ext::pair < T, R > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < typename T, typename R >
const std::string & xmlApi < ext::pair < T, R > >::xmlTagName ( ) {
	static std::string xmlTagName = "Pair";

	return xmlTagName;
}

template < typename T, typename R >
void xmlApi < ext::pair < T, R > >::compose ( ext::deque < sax::Token > & output, const ext::pair < T, R > & input ) {
	compose ( output, static_cast < const std::pair < T, R > > ( input ) );
}

template < typename T, typename R >
void xmlApi < ext::pair < T, R > >::compose ( ext::deque < sax::Token > & output, const std::pair < T, R > & input ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );

	alib::xmlApi < typename std::decay < T >::type >::compose ( output, input.first );
	alib::xmlApi < R >::compose ( output, input.second );

	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

template < typename T, typename R >
struct xmlApi < container::ObjectsPair < T, R > > {
	static container::ObjectsPair < T, R > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const container::ObjectsPair < T, R > & data );
};

template < typename T, typename R >
container::ObjectsPair < T, R > xmlApi < container::ObjectsPair < T, R > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	return container::ObjectsPair < T, R > ( xmlApi < ext::pair < T, R > >::parse ( input ) );
}

template < typename T, typename R >
bool xmlApi < container::ObjectsPair < T, R > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return xmlApi < ext::pair < T, R > >::first ( input );
}

template < typename T, typename R >
const std::string & xmlApi < container::ObjectsPair < T, R > >::xmlTagName ( ) {
	return xmlApi < ext::pair < T, R > >::xmlTagName ( );
}

template < typename T, typename R >
void xmlApi < container::ObjectsPair < T, R > >::compose ( ext::deque < sax::Token > & output, const container::ObjectsPair < T, R > & input ) {
	xmlApi < ext::pair < T, R > >::compose ( output, input );
}

template < >
struct normalize < std::pair < alib::Object, alib::Object > > {
	static std::pair < alib::Object, alib::Object > && eval ( std::pair < alib::Object, alib::Object > && source ) {
		return std::move ( source );
	}
};

template < class FirstType, class SecondType >
struct normalize < std::pair < FirstType, SecondType > > {
	static std::pair < alib::Object, alib::Object > eval ( std::pair < FirstType, SecondType > && source ) {
		return std::make_pair ( alib::Object ( alib::AnyObject < FirstType > ( std::move ( source.first ) ) ), alib::Object ( alib::AnyObject < SecondType > ( std::move ( source.second ) ) ) );
	}
};

template < >
struct normalize < ext::pair < alib::Object, alib::Object > > {
	static ext::pair < alib::Object, alib::Object > && eval ( ext::pair < alib::Object, alib::Object > && source ) {
		return std::move ( source );
	}
};

template < class FirstType, class SecondType >
struct normalize < ext::pair < FirstType, SecondType > > {
	static ext::pair < alib::Object, alib::Object > eval ( ext::pair < FirstType, SecondType > && source ) {
		return ext::make_pair ( alib::Object ( alib::AnyObject < FirstType > ( std::move ( source.first ) ) ), alib::Object ( alib::AnyObject < SecondType > ( std::move ( source.second ) ) ) );
	}
};

template < class FirstType, class SecondType >
struct normalize < container::ObjectsPair < FirstType, SecondType >, typename std::enable_if < ! std::is_same < container::ObjectsPair < FirstType, SecondType >, container::ObjectsPair < > >::value >::type > {
	static container::ObjectsPair < > eval ( container::ObjectsPair < FirstType, SecondType > && value ) {
		return container::ObjectsPair < > ( normalize < ext::pair < FirstType, SecondType > >::eval ( std::move ( value ) ) );
	}
};

} /* namespace alib */

#endif /* OBJECTS_PAIR_H_ */
