/*
 * ObjectsTree.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_TREE_H_
#define OBJECTS_TREE_H_

#include <tree>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>
#include <object/UniqueObject.h>
#include <object/AnyObject.h>

#include <core/normalize.hpp>

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occured.
 */
template < class ElementType >
class ObjectsTree final : public ext::tree < ElementType >, public ContainerBase {
public:
	explicit ObjectsTree ( ext::tree < ElementType > );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsTree & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef ObjectsTree < > normalized_type;
};

template < class ElementType >
ObjectsTree < ElementType >::ObjectsTree ( ext::tree < ElementType > raw ) : ext::tree < ElementType > ( std::move ( raw ) ) {
}

template < class ElementType >
ContainerBase * ObjectsTree < ElementType >::clone ( ) const {
	return new ObjectsTree ( * this );
}

template < class ElementType >
ContainerBase * ObjectsTree < ElementType >::plunder ( ) && {
	return new ObjectsTree ( std::move ( * this ) );
}

template < class ElementType >
int ObjectsTree < ElementType >::compare ( const ObjectsTree & other ) const {
	static ext::compare < ext::tree < ElementType > > comp;

	return comp ( static_cast < const ext::tree < ElementType > & > ( * this ), static_cast < const ext::tree < ElementType > & > ( other ) );
}

template < class ElementType >
void ObjectsTree < ElementType >::operator >>( std::ostream & os ) const {
	os << "(ObjectsTree " << static_cast < const ext::tree < ElementType > & > ( * this ) << ")";
}

template < class ElementType >
ObjectsTree < ElementType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return std::move ( ss ).str ( );
}

template < class ElementType >
alib::ObjectBase* ObjectsTree < ElementType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace alib {

template < typename T >
struct xmlApi < ext::tree < T > > {
	static ext::tree < T > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const ext::tree < T > & data );
};

template < typename T >
ext::tree < T > xmlApi < ext::tree < T > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "Root" );

	ext::tree < T > tree ( alib::xmlApi < T >::parse ( input ) );

	typename ext::tree < T >::const_children_iterator under = tree.root ( );

	unsigned level = 0;

	while ( sax::FromXMLParserHelper::isTokenType ( input, sax::Token::TokenType::START_ELEMENT ) ) {
		if ( sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, "Children" ) ) {
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "Children" );

			if ( level > 0 )
				under = std::prev ( under->end ( ) );

			++level;
		}

		tree.push_back ( under, alib::xmlApi < T >::parse ( input ) );

		while ( sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::END_ELEMENT, "Children" ) ) {
			sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "Children" );

			under = under->parent ( );
			--level;
		}
	}

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "Root" );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );
	return tree;
}

template < typename T >
bool xmlApi < ext::tree < T > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < typename T >
const std::string & xmlApi < ext::tree < T > >::xmlTagName ( ) {
	static std::string xmlTagName = "Tree";

	return xmlTagName;
}

template < typename T >
void xmlApi < ext::tree < T > >::compose ( ext::deque < sax::Token > & output, const ext::tree < T > & input ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	output.emplace_back ( "Root", sax::Token::TokenType::START_ELEMENT );

	unsigned level = 0;

	for ( typename ext::tree < T >::const_prefix_iterator iter = input.prefix_begin ( ); iter != input.prefix_end ( ); ) {
		while ( iter.getLevel ( ) > level ) {
			output.emplace_back ( "Children", sax::Token::TokenType::START_ELEMENT );
			++level;
		}

		alib::xmlApi < T >::compose ( output, * iter );
		++iter;

		while ( iter.getLevel ( ) < level ) {
			output.emplace_back ( "Children", sax::Token::TokenType::END_ELEMENT );
			--level;
		}
	}

	output.emplace_back ( "Root", sax::Token::TokenType::END_ELEMENT );
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

template < typename T >
struct xmlApi < container::ObjectsTree < T > > {
	static container::ObjectsTree < T > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const container::ObjectsTree < T > & data );
};

template < typename T >
container::ObjectsTree < T > xmlApi < container::ObjectsTree < T > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	return container::ObjectsTree < T > ( xmlApi < ext::tree < T > >::parse ( input ) );
}

template < typename T >
bool xmlApi < container::ObjectsTree < T > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return xmlApi < ext::tree < T > >::first ( input );
}

template < typename T >
const std::string & xmlApi < container::ObjectsTree < T > >::xmlTagName ( ) {
	return xmlApi < ext::tree < T > >::xmlTagName ( );
}

template < typename T >
void xmlApi < container::ObjectsTree < T > >::compose ( ext::deque < sax::Token > & output, const container::ObjectsTree < T > & input ) {
	return xmlApi < ext::tree < T > >::compose ( output, input );
}

template < >
struct normalize < ext::tree < alib::Object > > {
	static ext::tree < alib::Object > && eval ( ext::tree < alib::Object > && source ) {
		return std::move ( source );
	}
};

template < class ElementType >
struct normalize < ext::tree < ElementType > > {
	static ext::tree < alib::Object > eval ( ext::tree < ElementType > && source ) {
		ext::vector < ext::tree < alib::Object > > children;

		for ( ext::tree < ElementType > & child : source.getChildren ( ) ) {
			children.push_back ( normalize < ext::tree < ElementType > >::eval ( std::move ( child ) ) );
		}

		return ext::tree < alib::Object > ( alib::Object ( alib::AnyObject < ElementType > ( std::move ( source.getData ( ) ) ) ), std::move ( children ) );
	}
};

template < class ElementType >
struct normalize < container::ObjectsTree < ElementType >, typename std::enable_if < ! std::is_same < container::ObjectsTree < ElementType >, container::ObjectsTree < > >::value >::type > {
	static container::ObjectsTree < > eval ( container::ObjectsTree < ElementType > && value ) {
		return container::ObjectsTree < > ( normalize < ext::tree < ElementType > >::eval ( std::move ( value ) ) );
	}
};

} /* namespace alib */

#endif /* OBJECTS_TREE_H_ */
