/*
 * ObjectsTrie.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_TRIE_H_
#define OBJECTS_TRIE_H_

#include <trie>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>
#include <object/UniqueObject.h>
#include <object/AnyObject.h>

#include <core/normalize.hpp>

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occured.
 */
template < class KeyType, class ValueType >
class ObjectsTrie final : public ext::trie < KeyType, ValueType >, public ContainerBase {
public:
	explicit ObjectsTrie ( ext::trie < KeyType, ValueType > );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( ext::type_index ( typeid ( * this ) ) == ext::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return ext::type_index ( typeid ( * this ) ) - ext::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsTrie & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef ObjectsTrie < > normalized_type;
};

template < class KeyType, class ValueType >
ObjectsTrie < KeyType, ValueType >::ObjectsTrie ( ext::trie < KeyType, ValueType > raw ) : ext::trie < KeyType, ValueType > ( std::move ( raw ) ) {
}

template < class KeyType, class ValueType >
ContainerBase * ObjectsTrie < KeyType, ValueType >::clone ( ) const {
	return new ObjectsTrie ( * this );
}

template < class KeyType, class ValueType >
ContainerBase * ObjectsTrie < KeyType, ValueType >::plunder ( ) && {
	return new ObjectsTrie ( std::move ( * this ) );
}

template < class KeyType, class ValueType >
int ObjectsTrie < KeyType, ValueType >::compare ( const ObjectsTrie & other ) const {
	static ext::compare < ext::trie < KeyType, ValueType > > comp;

	return comp ( static_cast < const ext::trie < KeyType, ValueType > & > ( * this ), static_cast < const ext::trie < KeyType, ValueType > & > ( other ) );
}

template < class KeyType, class ValueType >
void ObjectsTrie < KeyType, ValueType >::operator >>( std::ostream & os ) const {
	os << "(ObjectsTrie " << static_cast < const ext::trie < KeyType, ValueType > & > ( * this ) << ")";
}

template < class KeyType, class ValueType >
ObjectsTrie < KeyType, ValueType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return std::move ( ss ).str ( );
}

template < class KeyType, class ValueType >
alib::ObjectBase* ObjectsTrie < KeyType, ValueType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace alib {

template < typename T, typename R >
struct xmlApi < ext::trie < T, R > > {
	static ext::trie < T, R > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const ext::trie < T, R > & data );

private:
	static ext::map < T, ext::trie < T, R > > parseChildren ( ext::deque < sax::Token >::iterator & input );
	static void composeChildren ( ext::deque < sax::Token > & out, const ext::map < T, ext::trie < T, R > > & children );
};

template < typename T, typename R >
ext::map < T, ext::trie < T, R > > xmlApi < ext::trie < T, R > >::parseChildren ( ext::deque < sax::Token >::iterator & input ) {
	ext::map < T, ext::trie < T, R > > children;

	while ( sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, "Child" ) ) {
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "Child" );

		T key = alib::xmlApi < T >::parse ( input );
		R value = alib::xmlApi < R >::parse ( input );
		ext::map < T, ext::trie < T, R > > innerChildren = parseChildren ( input );

		children.insert ( std::make_pair ( std::move ( key ), ext::trie < T, R > ( std::move ( value ), std::move ( innerChildren ) ) ) );

		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "Child" );
	}

	return children;
}

template < typename T, typename R >
ext::trie < T, R > xmlApi < ext::trie < T, R > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
	R value = alib::xmlApi < R >::parse ( input );
	ext::map < T, ext::trie < T, R > > children = parseChildren ( input );

	ext::trie < T, R > trie ( std::move ( value ), std::move ( children ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, xmlTagName ( ) );
	return trie;
}

template < typename T, typename R >
bool xmlApi < ext::trie < T, R > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < typename T, typename R >
const std::string & xmlApi < ext::trie < T, R > >::xmlTagName ( ) {
	static std::string xmlTagName = "Trie";

	return xmlTagName;
}

template < typename T, typename R >
void xmlApi < ext::trie < T, R > >::composeChildren ( ext::deque < sax::Token > & out, const ext::map < T, ext::trie < T, R > > & children ) {
	for ( const std::pair < const T, ext::trie < T, R > > & child : children ) {
		out.emplace_back ( "Child", sax::Token::TokenType::START_ELEMENT );

		alib::xmlApi < T >::compose ( out, child.first );
		alib::xmlApi < R >::compose ( out, child.second.getData ( ) );
		composeChildren ( out, child.second.getChildren ( ) );

		out.emplace_back ( "Child", sax::Token::TokenType::END_ELEMENT );
	}
}

template < typename T, typename R >
void xmlApi < ext::trie < T, R > >::compose ( ext::deque < sax::Token > & output, const ext::trie < T, R > & input ) {
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < R >::compose ( output, input.getData ( ) );
	composeChildren ( output, input.getChildren ( ) );
	output.emplace_back ( xmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

template < typename T, typename R >
struct xmlApi < container::ObjectsTrie < T, R > > {
	static container::ObjectsTrie < T, R > parse ( ext::deque < sax::Token >::iterator & input );
	static bool first ( const ext::deque < sax::Token >::const_iterator & input );
	static const std::string & xmlTagName ( );
	static void compose ( ext::deque < sax::Token > & output, const container::ObjectsTrie < T, R > & data );
};

template < typename T, typename R >
container::ObjectsTrie < T, R > xmlApi < container::ObjectsTrie < T, R > >::parse ( ext::deque < sax::Token >::iterator & input ) {
	return container::ObjectsTrie < T, R > ( xmlApi < ext::trie < T, R > >::parse ( input ) );
}

template < typename T, typename R >
bool xmlApi < container::ObjectsTrie < T, R > >::first ( const ext::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < typename T, typename R >
const std::string & xmlApi < container::ObjectsTrie < T, R > >::xmlTagName ( ) {
	return xmlApi < ext::trie < T, R > >::xmlTagName ( );
}

template < typename T, typename R >
void xmlApi < container::ObjectsTrie < T, R > >::compose ( ext::deque < sax::Token > & output, const container::ObjectsTrie < T, R > & input ) {
	xmlApi < ext::trie < T, R > >::compose ( output, input );
}

template < >
struct normalize < ext::trie < alib::Object, alib::Object > > {
	static ext::trie < alib::Object, alib::Object > && eval ( ext::trie < alib::Object, alib::Object > && source ) {
		return std::move ( source );
	}
};

template < class KeyType, class ValueType >
struct normalize < ext::trie < KeyType, ValueType > > {
	static ext::trie < alib::Object, alib::Object > eval ( ext::trie < KeyType, ValueType > && source ) {
		ext::map < alib::Object, ext::trie < alib::Object, alib::Object > > children;

		for ( std::pair < KeyType, ext::trie < KeyType, ValueType > > && child : ext::make_moveable_map ( source.getChildren ( ) ) ) {
			children.insert ( std::make_pair ( alib::Object ( alib::AnyObject < KeyType > ( std::move ( child.first ) ) ), normalizeRaw ( std::move ( child.second ) ) ) );
		}

		return ext::trie < alib::Object, alib::Object > ( alib::Object ( alib::AnyObject < ValueType > ( std::move ( source.getData ( ) ) ) ), std::move ( children ) );
	}
};

template < class KeyType, class ValueType >
struct normalize < container::ObjectsTrie < KeyType, ValueType >, typename std::enable_if < ! std::is_same < container::ObjectsTrie < KeyType, ValueType >, container::ObjectsTrie < > >::value >::type > {
	static container::ObjectsTrie < > eval ( container::ObjectsTrie < KeyType, ValueType > && value ) {
		return container::ObjectsTrie < > ( normalize < ext::trie < KeyType, ValueType > >::eval ( std::move ( value ) ) );
	}
};

} /* namespace alib */

#endif /* OBJECTS_TRIE_H_ */
